
# coding: utf-8

# # Classification du Raclette à la Tensor
# 
# > *Raclette is a dish indigenous to parts of Switzerland. The raclette cheese round is heated, either in front of a fire or by a special machine, then scraped onto diners' plates; the term raclette derives from the French word racler, meaning "to scrape", a reference to the fact that the melted cheese must be scraped from the unmelted part of the cheese onto the plate.* ([Wikipedia](https://en.wikipedia.org/wiki/Raclette))
# 
# This notebook created in clear imitation of [Not Hotdog](https://medium.com/@timanglade/how-hbos-silicon-valley-built-not-hotdog-with-mobile-tensorflow-keras-react-native-ef03260747f3) at the [Open Food Data Hackdays](https://food.opendata.ch) / [Applied Machine Learning Days](https://www.appliedmldays.org/), and was based on code and tips from:
# 
# - https://github.com/xpzouying/animals-classification (mostly)
# - http://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html
# - https://github.com/IMG-PRCSNG/dlexperiments
# - http://pytorch.org/docs/master/tensors.html
# 
# ..the underlying model used is ResNet-18, described here:
# 
# - https://arxiv.org/abs/1512.03385
# - https://www.pyimagesearch.com/2017/03/20/imagenet-vggnet-resnet-inception-xception-keras/
# 
# ..and this was also motivated by this tweet:
# 
# <blockquote class="twitter-tweet" data-partner="tweetdeck"><p lang="en" dir="ltr">Having troubles finding a good &amp; reliable reference to add in <a href="https://twitter.com/hashtag/wikidata?src=hash&amp;ref_src=twsrc%5Etfw">#wikidata</a> that wd:Q20748 (Raclette) has country of origin Switzerland. It currently says France. 🤔Anyone a good pointer? <a href="https://twitter.com/hashtag/opendatach?src=hash&amp;ref_src=twsrc%5Etfw">#opendatach</a> <a href="https://twitter.com/hashtag/switzerland?src=hash&amp;ref_src=twsrc%5Etfw">#switzerland</a></p>&mdash; Cristina Sarasua (@csarasuagar) <a href="https://twitter.com/csarasuagar/status/956970996292050945?ref_src=twsrc%5Etfw">January 26, 2018</a></blockquote>
# 
# For an introduction to PyTorch and Tensors, [start here](http://pytorch.org/tutorials/beginner/blitz/tensor_tutorial.html#sphx-glr-beginner-blitz-tensor-tutorial-py) (pytorch.org)
# 
# To skip right to the juicy part, [click here](#testing).

# In[2]:


import torch
import torch.nn as nn
from torch.autograd import Variable
import torchvision
from torch.utils.data import DataLoader
from torchvision import transforms
from torchvision.datasets import ImageFolder
import matplotlib.pyplot as plt
from PIL import Image


# Experiencing a problem here? You probably need to `conda install` or `pip install` some more lovely packages.

# In[3]:


data_dir = "data"


# Create a `data` folder with subfolders, each containing JPG images, e.g.:
# 
# ```
# data/raclette/image_1.jpg
# data/not_raclette/image_2.jpg
# ```
# 
# The subfolder name (e.g. `raclette`) will correspond to our **label**. You can pick test images from the folder [here](#testing).
# 
# Here the main set up work takes place, depending on how many images you have `ImageFolder` may take a while:

# In[4]:


# dataset pre-processing
train_transform = transforms.Compose([
    transforms.CenterCrop(200), # crop our images to a 200 square
    transforms.Resize(224), # resize down to 224x224 (power of 2)
    transforms.ToTensor(), # convert to a Tensor
])

# details on how to use ImageFolder in https://github.com/pytorch/vision#imagefolder
tset = ImageFolder(data_dir, transform=train_transform)
train_dataloader = DataLoader(tset, batch_size=4, shuffle=True)
num_classes = len(set(tset.classes))

# use pretrained resnet18 network with image input
model = torchvision.models.resnet18(pretrained=True)
num_ftrs = model.fc.in_features # e.g. 2048
model.fc = torch.nn.Linear(num_ftrs, num_classes)
print("Features: ", num_ftrs)

# run in cuda
if torch.cuda.is_available():
    model = model.cuda()
print(model)


# You may ponder the runes above for a while, which explain the model that was just constructed.
# 
# Now we are ready to load the data.

# In[5]:


# optimizer
optimizer = torch.optim.SGD(model.parameters(), lr=0.001, momentum=0.9)

# loss function
loss_fn = nn.CrossEntropyLoss()

# begin to train models
model.train()

# getting an error here? install the latest torchvision
# pip install https://github.com/pytorch/vision/archive/master.zip
for inputs, labels in train_dataloader:
    if torch.cuda.is_available():
        # use cuda
        inputs = Variable(inputs.cuda())
        labels = Variable(labels.cuda())
    else:
        # use cpu
        inputs = Variable(inputs)
        labels = Variable(labels)

    # zero the parameter gradients
    optimizer.zero_grad()

    # forward
    outputs = model(inputs)
    loss = loss_fn(outputs, labels)
    loss.backward()
    optimizer.step()


# At this point it may be wise to go and brew yourself a hot cup of herbal tea from the Swiss alps.

# In[6]:


model.eval()


# In[7]:


def test_image(image_file):
    plt.imshow(plt.imread(image_file))

    img2 = Image.open(image_file)
    img2 = train_transform(img2)
    img2 = img2.unsqueeze(0)  # pytorch only accepts batch images

    img2 = Variable(img2)
    if torch.cuda.is_available(): img2 = img2.cuda()

    target = model(img2)

    # get the predict classes
    _, pred = torch.max(target.data, 1)

    print(nn.functional.softmax(target))
    print("It must be: ", tset.classes[pred[0][0]], " !!!")


# ## Testing
# <a name="testing"></a>
# 
# With a handy function we can throw a bunch of images (from outside of the training set) at the classifier, and see what happens.

# In[12]:


test_image("data/test/r1.jpg")


# Source: [Raclette-Negative](https://commons.wikimedia.org/wiki/File:Raclette-Negative0-03-1A%281%29.jpg) CC BY-SA 2.0 France [Rama](https://commons.wikimedia.org/wiki/User:Rama)

# In[13]:


test_image("data/test/n1.jpg")


# Source: [Burger King A.1. Steakhouse XT](https://www.flickr.com/photos/93512023@N00/4425762959) CC BY-SA 2.0 [Jason Lam](https://www.flickr.com/photos/mesohungry/)

# In[14]:


test_image("data/test/r2.jpg")


# Source: [Raclette2](https://commons.wikimedia.org/wiki/File:Raclette2.jpg) CC BY-SA 3.0 [Grcampbell](https://commons.wikimedia.org/wiki/File:Raclette2.jpg)

# In[15]:


test_image("data/test/n2.jpg")


# Source: [Hot dog with baked beans and potato salad](https://www.flickr.com/photos/preppybyday/5076302063/) CC BY-SA 2.0 [TheCulinaryGeek](https://www.flickr.com/photos/preppybyday/)

# In[16]:


test_image("data/test/r3.jpg")


# Source: [Christmas Raclette](https://www.flickr.com/photos/kentwang/16218650136/in/photolist-qHbPCU-66nHKx-5NkktB-7PsgA-atjbb-dG6m42-4kDu1a-7t9Dxx-4Gw9Qf-5moFTv-79gHq6-9qEGm1-dKraYB-8AtmHk-dGUsFZ-4EYzai-cB5u75-kepiaQ-4H188F-e3ba8D-7Jndvq-7WVdun-7WYr1Y-5Tk4xH-9wRgzo-98ZWDe-7PmTC7-4H5ikE-Hc8WS-98ZXfa-4BQfm-5LNsjZ-7umLSH-7PhNmi-74fT14-7KZoeF-HcaNk-7KZok6-7PmUbW-98ZX7i-6CsTDn-H5Rfy-7PhRgt-Hc8Z1-DmGteC-HcaWV-7PmQPC-8Ez1E-Ru94YL-2kr82) CC BY-SA 2.0 [Kent Wang](https://www.flickr.com/photos/kentwang/)

# In[18]:


test_image("data/test/n3.jpg")


# Source: [5 a day salad](https://commons.wikimedia.org/wiki/File:5aday_salad.jpg) Public Domain

# Well, this was fun! Thanks to everyone at the Hackdays/EPFL for an awesome event.
